%h1 Extending C with Python

%p
  Sometimes it is useful to borrow functionality that is implemented in a Python
  library. To do this we can initialize an embedded Python interpreter, and then
  load functions.

%h2 Getting Started

%p
  The basic structure of initializing the interpreter and loading the
  %a{:href=>"https://www.nsc.liu.se/~kent/python-hostlist/"} python-hostlist
  module is as follows

:codeblock
  :::c
  #include <Python.h>

  int
  main(int argc, char *argv[])
  {
      PyObject *pModule;

      Py_SetProgramName(argv[0]);
      Py_Initialize();

      /* import module */
      pModule = PyImport_ImportModule("hostlist");
      if (pModule == NULL) {
          if (PyErr_Occurred())
              PyErr_Print();
          exit(1);
      }

      Py_DECREF(pModule);
      Py_Finalize();
      return 0;
  }

%p
  Now we can build using

:codeblock
  :::make
  CFLAGS != python2.7-config --cflags
  LDFLAGS != python2.7-config --ldflags

  hostlist: hostlist.c
      cc ${CFLAGS} -L/usr/local/lib ${LDFLAGS} $< -o $@

%h2 Calling Functions

%p
  Python arguments are return values are all
  %code PyOjbect
  pointers, and we can use the well-documented set of
  %a{:href=>"https://docs.python.org/3/c-api/"} standard functions
  to instantiate data and validate/convert outputs

:codeblock
  :::c
  PyObject* expand_hostlist(PyObject *pModule, PyObject* pHostlistExp) {
      PyObject *pFunc;
      PyObject *pArgs, *pValue;

      /* get pointer to function */
      pFunc = PyObject_GetAttrString(pModule, "expand_hostlist");
      if (!pFunc) {
          if (PyErr_Occurred())
              PyErr_Print();
          exit(1);
      }

      /* construct an argument list */
      pArgs = PyTuple_New(1);
      PyTuple_SetItem(pArgs, 0, pHostlistExp);
      pValue = PyObject_CallObject(pFunc, pArgs);

      Py_DECREF(pArgs);
      Py_DECREF(pFunc);

      /* ensure the return value is what we expect */
      if (pValue && PyList_Check(pValue)) {
          return pValue;
      }
      return NULL;
  }

%p
  To call this function, simply convert a C string into a Python object using
  = succeed '.' do
    %code PyString_FromString(argv[1])
